% Computational Intelligence: a logical approach. 
% Prolog Code.
% PARTIAL ORDER PLANNER using STRIPS REPRESENTATION (Figures 8.5 & 8.6)
% Copyright (c) 1998, Poole, Mackworth, Goebel and Oxford University Press.

% Definitions:
% An agenda is a list of goals.
% A goal is of the form goal(P,NA) where P is a proposition 
%   that is a precondition of action instance index NA.
% A plan is of the form plan(As,Os,Ls) where 
%   As is a list of actions instances,
%      An action instance is of the form act(N,A) where N is an integer 
%      and A is an action (this is needed in case there are multiple 
%      instances of the same action). N is the action instance index.
%   Os is a list of ordering constraints (A1<A2) where A1 & A2 
%      are action instance indexes.
%   Ls is a list of causal links.
% A causal link is of the form cl(A0,P,A1) where A0 and A1 are action 
%   instance indexes, and P is a proposition that is a precondition of 
%   action A1 means that A0 is making P true for A1.

:- op(1200,xfx,[<-]).
% N.B. we assume that conjunctions are represented as lists.
% `\=' is the object level not equal.
:- op(700,xfx, \=).


:- op(700,xfx, iss).
:- op(700,xfx, ==?).
:- op(700,xfx, <?).
:- op(700,xfx, >?).

% pop(CPlan, Agenda,DIC, FPlan,DB) is true if
%  CPlan is the current plan
%  Agenda is the current agenda
%  FPlan is the final plan (with all subgoals supported)
%  and there are DB or fewer actions in the plan (i.e., DB is the depth-bound).

pop(Plan,[],Plan,_,_,_):-!.

pop(CPlan,Agenda,FPlan,ID,DB,MaxDB) :-
   select2(goal(G,GG),Agenda,MyAgenda,10),!,
   updateDB(Agenda,DB,MaxDB,DB2),
   (G <- B),
   makeintogoal(B,GG,MyAgenda,NR),
   select(Goal,NR,Agenda1),
%   writeln('checking: '), writeln(Goal), writeln(Agenda1),writeln(CPlan),flush,
   solve_goal(Goal,CPlan,NPlan,Agenda1,NAgenda,ID,NID,DB2,NDB),
%   writeln('success: '), writeln(Goal), writeln(NAgenda),writeln(NPlan),
%   writeln(''),flush,
   pop(NPlan,NAgenda,FPlan,NID,NDB,MaxDB).

pop(CPlan,Agenda,FPlan,ID,DB,MaxDB) :-
   select(Goal,Agenda,Agenda1),
   updateDB(Agenda,DB,MaxDB,DB2),
%   writeln('checking: '), writeln(Goal), writeln(Agenda1),writeln(CPlan),
   solve_goal(Goal,CPlan,NPlan,Agenda1,NAgenda,ID,NID,DB2,NDB),
%   writeln('success: '), writeln(Goal), writeln(NAgenda),writeln(NPlan),
%   writeln(''),
   pop(NPlan,NAgenda,FPlan,NID,NDB,MaxDB).

updateDB([goal(_G,GG)|_A],_DB,MaxDB,MaxDB) :- GG==finish,!.
updateDB(_A,DB,_MaxDB,DB).

% select(Goal,Agenda,NewAgenda) is true if Goal is selected from Agenda with
%    NewAgenda the remaining elements.

select2(_,_,_,0):-!,fail.
select2(goal(G,GG),[goal(G,GG)|R],R,_ID):-
                 G= adjacent(point(Y,Z),point(T,W),_Jahat),number(Y),
                 number(Z),number(T),number(W),!.
select2(goal(G,GG),[goal(G,GG)|R],R,_ID):-
                 G= adjacent(point(Y,Z),point(_T,_W),Jahat),number(Y),
                 number(Z),atom(Jahat),!.
select2(goal(G2,GG),[goal(G,GG)|R],R,_ID):-
                 G= adjacent(point(Y,Z),point(T,W),Jahat),atom(Jahat),
                 number(T),number(W),!,
                 G2=adjacent2(point(T,W),point(Y,Z),Jahat).
select2(Return,[X|A],[X|B],ID):-NID is ID-1,select2(Return,A,B,NID),!.

select(goal(G,GG),[goal(G,GG)|A],A) :-
   primitive(G),!.
select(goal((n(N)),GG),[goal((n(N)),GG)|A],A) :- !.
select(goal((m(M)),GG),[goal((m(M)),GG)|A],A) :- !.
select(goal((X \= Y),GG),[goal((X \= Y),GG)|A],A) :- !.
select(goal((X ==? Y),GG),[goal((X ==? Y),GG)|A],A) :- !.
select(goal((X iss Y),GG),[goal((X iss Y),GG)|A],A) :- !.
select(goal((X <? Y),GG),[goal((X <? Y),GG)|A],A) :- !.
select(goal((X >? Y),GG),[goal((X >? Y),GG)|A],A) :- !.
select(P,[goal(G,GG)|R],A) :-
   (G <- B),
   makeintogoal(B,GG,R,NR),
   select(P,NR,A).

makeintogoal([],_,R,R).
makeintogoal([H|T],GG,R,[goal(H,GG)|NR]) :-
   makeintogoal(T,GG,R,NR).

% x_goal(Goal,CPl,NPl,CAg,NAg,DB,DB1)
%    chooses an action to
%    solve Goal, updating plan CPl to NPl and agenda CAg to NAg,
%    and updating depth bound DB to depth bound DB1, and


solve_goal(goal((n(N)),_),Plan,Plan,Agenda,Agenda,ID,ID,DB,DB) :-
    !, n(N).
solve_goal(goal((m(M)),_),Plan,Plan,Agenda,Agenda,ID,ID,DB,DB) :-
    !, m(M).
solve_goal(goal((X ==? Y),_),Plan,Plan,Agenda,Agenda,ID,ID,DB,DB) :-
    !,
    X=Y.
solve_goal(goal((X >? Y),_),Plan,Plan,Agenda,Agenda,ID,ID,DB,DB) :-
    !,
    X>Y.
solve_goal(goal((X <? Y),_),Plan,Plan,Agenda,Agenda,ID,ID,DB,DB) :-
    !,
    X<Y.
    
solve_goal(goal((X iss Y),_),Plan,Plan,Agenda,Agenda,ID,ID,DB,DB) :-
    !, \+ var(Y),
    X is Y.
    
% Heuristic 1 Check
solve_goal(goal(stone_at(_Number, _P),_),Plan,Plan,Agenda,Agenda,ID,ID,DB,DB) :-
   heur1_check(Agenda, MaxDB),(ID < MaxDB), !, fail.
   
% CASE 1: use existing action
solve_goal(goal(P,A1),plan(As,Os,Ls),plan(As,NOs,[cl(N0,P,A1)|Ls]),Ag,Ag,ID,ID,DB,DB) :-
   member(act(N0,Act0),As),
   achieves(Act0,P),
   add_constraint(N0<A1,Os,Os1) ,
   incorporate_causal_link(cl(N0,P,A1),As,Os1,NOs,A1).


% CASE 2: add new action. 
%   Note that DB acts as the unique index of the new action instance.
solve_goal(goal(P,A1),
           plan(As,Os,Ls),
           plan([act(ID,Act0)|As],NOs,[cl(ID,P,A1)|Ls]),Ag,NAg,ID,NID,DB,NDB) :-
   DB>0,
   ID>0,
   achieves(Act0,P),
   add_constraint(ID<A1,Os,Os1),
   add_constraint(start<ID,Os1,Os2),
   incorporate_action(act(ID,Act0),Ls,Os2,Os3,A1),
   incorporate_causal_link(cl(ID,P,A1),As,Os3,NOs,A1),
   add_preconds(act(ID,Act0),Ag,NAg),
%   writeln(['*** new action ',act(DB,Act0), ' to achieve ',P,' for ',A1]),
   NDB is DB-1,
   NID is ID-1.


% add_constraint(A0<A1,Os,Os1) adds ordering constraint A0<A1 to 
%    partial ordering Os producing partial ordering Os1.
% Fails if A0<A1 is inconsistent with Os.
% We represent parial orderings as their transitive closure.
add_constraint(C,L,L) :- member(C,L),!.      % green cut for efficiency only
add_constraint(A0<A1,L1,L2) :-
   different(A0,A1),
%   \+ member(A0<A1,L1),    % ommitted because of the cut for efficiency.
   \+ member(A1<A0,L1),
%   writeln([' ... adding constraint ',A0<A1]),
   add_constraint1(A0<A1,L1,L1,L2).

% add_constraint1(A0<A1,Os,AOs,Os1) adds ordering constraint A0<A1 
%    partial ordering Os is the remaining orderings to be checked
%    AOs is the list of all orderings
%    Os1 is the final set of orederings
add_constraint1(A0<A1,[],AOs,NOs) :-
    insert(A0<A1,AOs,NOs).
add_constraint1(A0<A1,[A1<A2|R],AOs,NR) :-
   different(A0,A2),
   insert(A0<A2,AOs,AOs1),
   add_constraint1(A0<A1,R,AOs1,NR).
add_constraint1(A0<A1,[A2<A0|R],AOs,NR) :-
   different(A1,A2),
   insert(A2<A1,AOs,AOs1),
   add_constraint1(A0<A1,R,AOs1,NR).
add_constraint1(A0<A1,[A2<A3|R],AOs,NR) :-
   different(A0,A3),
   different(A1,A2),
   add_constraint1(A0<A1,R,AOs,NR).

% incorporate_causal_link(CL, As, Os, NOs)
%  incororates causal link CL to links Ls producing new links NLs, and updating
%  parial ordering Os to NOs. RAs is the list of remaining actions.
%  As is the list of all actions.
incorporate_causal_link(CL, As, Os, NOs,Niyaz) :-
%   writeln(['Adding causal link: ',CL,nl,'  As=',As,nl,'  Os=',Os]),
   incorporate_causal_link1(CL, As,Os,NOs,Niyaz)
%,   writeln([' Added causal link: ',CL,nl,'  As=',As,nl,'  Os=',NOs])
.
incorporate_causal_link1(_,[],Os,Os,_Niyaz).
incorporate_causal_link1(CL,[A|RAs],Os,NOs,Niyaz) :-
   protect(CL,A,Os,Os1,Niyaz),
   incorporate_causal_link1(CL,RAs,Os1,NOs,Niyaz).

incorporate_action(_,[],Os,Os,_Niyaz).
incorporate_action(A,[CL|Ls],Os,NOs,Niyaz) :-
   protect(CL,A,Os,Os1,Niyaz),
   incorporate_action(A,Ls,Os1,NOs,Niyaz).

% protect(Cl,Action,Os0,Os1) protects causal link CL from Action if necessary
protect(cl(A0,_,_),act(A0,_),Os,Os,_Niyaz) :- !.
protect(cl(A0,_,_),act(NA,_),Os,Os,_Niyaz) :-
   member(NA<A0,Os),!.
protect(cl(_,_,A1),act(A1,_),Os,Os,_Niyaz) :- !.
protect(cl(_,_,A1),act(NA,_),Os,Os,_Niyaz) :-
   member(A1<NA,Os), !.
protect(cl(_,P,_),act(_,A),Os,Os,_Niyaz) :-
%  different(NA,A0), \+ member(NA<A0,Os), different(NA,A1), \+ member(A1<NA,Os), 
     % deleted because of cuts
%   when(ground((A,P)), \+ deletes(A,P)).
    \+ deletes(A,P),!. %CHANGED CODE
protect(cl(A0,_P,A1),act(NA,_A),Os,Os1,Niyaz) :-
%  different(NA,A0), \+ mexmber(NA<A0,Os), different(NA,A1), \+ member(A1<NA,Os),
     % deleted because of cuts
%   deletes(A,P), %CHANGED CODE
   enforce_order(A0,NA,A1,Os,Os1,Niyaz).

%demote if the action needing this action isn't finish
enforce_order(A0,A,_,Os,Os1,Niyaz) :-(Niyaz\==finish),
   add_constraint(A<A0,Os,Os1)
%,   writeln(['Putting ',A,' before ',A0])
.
enforce_order(_,A,A1,Os,Os1,Niyaz) :-(Niyaz\==finish),!,
   add_constraint(A1<A,Os,Os1)
%,   writeln(['Putting ',A,' after ',A1])
.
%promote if finish
enforce_order(_,A,A1,Os,Os1,finish) :-
   add_constraint(A1<A,Os,Os1)
%,   writeln(['Putting ',A,' after ',A1])
.
%demote has a lower priority than promote if we are in finish
enforce_order(A0,A,_,Os,Os1,finish) :-!,
   add_constraint(A<A0,Os,Os1)
%,   writeln(['Putting ',A,' before ',A0])
.

consistent(A0<A1,Os) :-
   different(A0,A1),
   \+ member(A1<A0,Os).

% =====================================
% interface to STRIPS notation
/* we assume now that these actions are defined via new-strips
achieves(A,P) :-
   addlist(A,AL),
   member(P,AL).
achieves(init,P) :-
   holds(P,init).
deletes(A,P) :-
   deletelist(A,AL),
   member(P,AL).
*/
add_preconds(act(N,A),Ag0,Ag1) :-
   preconditions(A,Ps),
%   add_these_preconds(Ps,N,Ag0,Ag1), % makes the agenda a stack
   add_these_preconds(Ps,N,[],NPs),
   append(NPs,Ag0,Ag1).              % makes the agenda a queue
add_these_preconds([],_,Ag,Ag).
%add_these_preconds([X \= Y|R],A,Ag0,Ag1) :- !,
%   different(X,Y),
%   add_these_preconds(R,A,Ag0,Ag1).
add_these_preconds([P|R],A,Ag0,[goal(P,A)|Ag1]) :-
%   primitive(P),
   add_these_preconds(R,A,Ag0,Ag1).
%add_these_preconds([P|R],A,Ag0,Ag2) :-
%   \+ primitive(P),
%   (P <- B),
%   add_these_preconds(B,A,Ag0,Ag1),
%   add_these_preconds(R,A,Ag1,Ag2).

solve(Goals,Plan,DB) :-
   add_these_preconds(Goals,finish,[],Ag),
   pop(plan([act(finish,end),act(start,init)],[start<finish],[]),Ag,Plan,17,DB,DB),!.
solve(Goals,Plan,DB):-
not(DB==15),
   write('we are going in DB: '),NDB is DB+1,writeln(NDB),solve(Goals,Plan,NDB).
% seq(Plan,Seq) extracts a legal sequence Seq from Plan
seq(plan([],_,_),[]).
seq(plan(As,Os,_),[A|P]) :-
   remove(act(N,A),As,As1),
   \+ (member(act(N1,_),As1), member(N1<N,Os)),
   seq(plan(As1,Os,_),P).

% =============================================================================

different(X,Y) :-
   X \== Y.
   
my_different(X,Y,DIC) :-
   X=..A, Y=..B, my_different2(A,B,DIC), !.

my_different2([],[],_DIC) :- !,fail.
my_different2([AH|AT],[BH|BT],_DIC) :-
   nonvar(AH), nonvar(BH), !,AH \== BH, my_different2(AT,BT).
my_different2([AH|_AT],[BH|_BT],DIC) :-
   member((AH\=BH),DIC),!.
my_different2([AH|_AT],[BH|_BT],DIC) :-
   member((BH\=AH),DIC),!.

append([],L,L).
append([H|X],Y,[H|Z]) :-
   append(X,Y,Z).

% member(X,L) is true if X is a member of list L
member(X,[X|_]).
member(X,[_|L]) :-
   member(X,L).

% remove(X,L,R) is true if X is a member of list L, with remaining elements R
remove(X,[X|R],R).
remove(X,[A|L],[A|R]) :-
   remove(X,L,R).

% insert(E,L,L1) inserts E into L producing L1 
% E is not added it is already there.
insert(X,[],[X]).
insert(A,[A|R],[A|R]).
insert(A,[B|R],[B|R1]) :-
   different(A,B),
   insert(A,R,R1).

% writeln(L) is true if L is a list of items to be written on a line, followed by a newline.
writelnw(L) :- \+ \+ (numbervars(L,0,_), writelnw(L) ).
writelnw([]) :- nl.
writelnw([nl|T]) :- !,nl, writeln(T).
writelnw([H|T]) :- write(H), writeln(T).

% =============================================================================

% TRY THE FOLLOWING QUERIES with delrob_strips.pl:
% solve([carrying(rob,k1)],P,3), seq(P,S).
% solve([sitting_at(k1,lab2)],P,7), seq(P,S).
% solve([carrying(rob,parcel),sitting_at(rob,lab2)],P,9), seq(P,S).
% solve([sitting_at(rob,lab2),carrying(rob,parcel)],P,9), seq(P,S).
